from django.core.checks import Warning, register


default_app_config = 'reactor.app.Reactor'


@register()
def check_for_turbolinks_middleware(app_configs, **kwargs):
    from . import settings
    from django.conf import settings as dj_settings

    reactor_middleware = 'reactor.middleware.turbolinks_middleware'
    problem = (
        settings.INCLUDE_TURBOLINKS and
        reactor_middleware not in dj_settings.MIDDLEWARE
    )
    if problem:
        return [Warning(
            'Turbolinks middleware is missing',
            hint=(
                f'When you set REACTOR_INCLUDE_TURBOLINKS, during the '
                f'redirects the URL of the browser could be that is not up to '
                f'date. To avoid that include `{reactor_middleware}` in the '
                f'`MIDDLEWARE`.'
            ),
            id='reactor.W001',
        )]
    else:
        return []



from typing import Generator
import orjson

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models


def loads(text_data):
    return orjson.loads(text_data)


def dumps(obj, indent=False):
    option = (
        orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY
    )
    if indent:
        option |= orjson.OPT_INDENT_2
    return orjson.dumps(obj, default=default, option=option).decode()


def default(o):
    if isinstance(o, models.Model):
        return o.pk

    if isinstance(o, models.QuerySet):
        return list(o.values_list('pk', flat=True))

    if isinstance(o, (Generator, set)):
        return list(o)

    if hasattr(o, '__json__'):
        return o.__json__()

    return DjangoJSONEncoder().default(o)




from django.dispatch import receiver
from django.db import models
from django.db.models.signals import post_save, pre_delete, m2m_changed

from .utils import broadcast
from .settings import AUTO_BROADCAST


MODEL = AUTO_BROADCAST.get('MODEL', False)
MODEL_PK = AUTO_BROADCAST.get('MODEL_PK', False)
RELATED = AUTO_BROADCAST.get('RELATED', False)
M2M = AUTO_BROADCAST.get('M2M', False)


if MODEL or MODEL_PK or RELATED:
    @receiver(post_save)
    def broadcast_post_save(sender, instance, created=False, **kwargs):
        name = sender._meta.model_name
        if MODEL:
            broadcast(name)
            created and broadcast(f'{name}.new')

        if instance.pk is not None:
            MODEL_PK and broadcast(f'{name}.{instance.pk}')
            RELATED and broadcast_related(sender, instance, created=created)

    @receiver(pre_delete)
    def broadcast_post_delete(sender, instance, **kwargs):
        name = sender._meta.model_name
        MODEL and broadcast(name)

        if instance.pk is not None:
            MODEL and broadcast(f'{name}.del')
            MODEL_PK and broadcast(f'{name}.{instance.pk}')
            RELATED and broadcast_related(sender, instance, deleted=True)


def broadcast_related(sender, instance, deleted=False, created=False):
    for field in get_related_fields(sender):
        if field['is_m2m']:
            fk_ids = getattr(instance, field['name']).values_list(
                'id', flat=True
            )
        else:
            fk_ids = filter(None, [getattr(instance, field['name'])])

        if fk_ids:
            group_names = [
                f'{field["related_model_name"]}.{fk_id}.{field["related_name"]}'
                for fk_id in fk_ids
            ]
            broadcast(*group_names)
            if created:
                broadcast(*[f'{gn}.new' for gn in group_names])
            if deleted:
                broadcast(*[f'{gn}.del' for gn in group_names])


MODEL_RELATED_FIELDS = {}


def get_related_fields(model):
    related_fields = MODEL_RELATED_FIELDS.get(model)
    if related_fields is None:
        fields = []
        for field in model._meta.get_fields():
            if isinstance(field, (models.ForeignKey, models.ManyToManyField)):
                related_name = field.related_query_name()
                if related_name != '+':
                    is_m2m = isinstance(field, models.ManyToManyField)
                    if not is_m2m or M2M and is_m2m:
                        fields.append({
                            'is_m2m': is_m2m,
                            'name': field.attname,
                            'related_name': related_name,
                            'related_model_name':
                                field.related_model._meta.model_name,
                        })
        related_fields = MODEL_RELATED_FIELDS[model] = tuple(fields)
    return related_fields


if M2M:
    @receiver(m2m_changed)
    def broadcast_m2m_changed(
        sender, instance, action, model, pk_set, **kwargs
    ):
        if action.startswith('post_') and instance.pk:
            model_name = model._meta.model_name
            attr_name = get_name_of(sender, model)
            updates = [f'{model_name}.{pk}.{attr_name}' for pk in pk_set or []]
            broadcast(*updates)
            broadcast(*[f'{u}.{instance.pk}' for u in updates])

            model = type(instance)
            model_name = model._meta.model_name
            attr_name = get_name_of(sender, model)
            update = f'{model_name}.{instance.pk}.{attr_name}'
            broadcast(update)
            broadcast(*[f'{update}.{pk}' for pk in pk_set or []])


def get_name_of(through, model):
    for model_field in model._meta.get_fields():
        found = (
            getattr(model_field, 'through', None) or
            getattr(getattr(model, model_field.name, None), 'through', None)
        )
        if through is found:
            return model_field.name




import django

django.setup(set_prefix=False)


def get_asgi_application():
    from django.conf import settings
    from django.utils.module_loading import import_module
    from channels.auth import AuthMiddlewareStack
    from channels.routing import ProtocolTypeRouter, URLRouter

    urls = import_module(settings.ROOT_URLCONF)
    return ProtocolTypeRouter({
        'websocket': AuthMiddlewareStack(URLRouter(
            urls.websocket_urlpatterns,
        ))
    })



import logging

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer

from . import json
from .component import RootComponent, Component


log = logging.getLogger('reactor')


class ReactorConsumer(JsonWebsocketConsumer):
    channel_name = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subscriptions = set()

    # Group operations

    def subscribe(self, event):
        room_name = event['room_name']
        if room_name not in self.subscriptions:
            log.debug(f':: SUBSCRIBE {self.channel_name} {room_name}')
            async_to_sync(self.channel_layer.group_add)(
                room_name,
                self.channel_name
            )
            self.subscriptions.add(room_name)

    def unsubscribe(self, event):
        room_name = event['room_name']
        if room_name in self.subscriptions:
            log.debug(f':: UNSUBSCRIBE {self.channel_name} {room_name}')
            async_to_sync(self.channel_layer.group_discard)(
                room_name,
                self.channel_name
            )
            self.subscriptions.discard(room_name)

    # Channel events

    def connect(self):
        super().connect()
        self.scope['channel_name'] = self.channel_name
        self.root_component = RootComponent(context=self.scope)
        self.send_json({
            'type': 'components',
            'component_types': {
                name: c.extends for name, c in Component._all.items()
            }
        })
        log.debug(f':: CONNECT {self.channel_name}')

    def disconnect(self, close_code):
        for room in list(self.subscriptions):
            self.unsubscribe({'room_name': room})
        log.debug(f':: DISCONNECT {self.channel_name}')

    # Dispatching

    def receive_json(self, request):
        name = request['command']
        payload = request['payload']
        getattr(self, f'receive_{name}')(**payload)

    def receive_join(self, tag_name, state):
        log.debug(f'>>> JOIN {tag_name} {state}')
        component = self.root_component.get_or_create(tag_name, **state)
        html_diff = component._render_diff()
        self.render({'id': component.id, 'html_diff': html_diff})

    def receive_user_event(self, id, name, args):
        log.debug(f'>>> USER_EVENT {name} {args}')
        html_diff = self.root_component.dispatch_user_event(id, name, args)
        self.render({'id': id, 'html_diff': html_diff})

    def receive_leave(self, id):
        self.root_component.pop(id)

    # Internal event

    def update(self, event):
        log.debug(f'>>> UPDATE {event}')
        for render_event in self.root_component.propagate_update(event):
            self.render(render_event)

    def send_component(self, event):
        log.debug(f'>>> DISPATCH {event}')
        self.receive_user_event(
            event['state']['id'],
            event['name'],
            event['state']
        )

    # Broadcasters

    def render(self, event):
        if event['html_diff']:
            log.debug(f"<<< RENDER {event['id']}")
            self.send_json(dict(event, type='render'))

    def remove(self, event):
        log.debug(f"<<< REMOVE {event['id']}")
        self.receive_leave(event['id'])
        self.send_json(dict(event, type='remove'))

    def redirect(self, event):
        log.debug(f"<<< REDIRECT {event['url']}")
        self.send_json(dict(event, type='redirect'))

    def push_state(self, event):
        log.debug(f"<<< PUSH-STATE {event['url']}")
        self.send_json(dict(event, type='push_state'))

    @classmethod
    def decode_json(cls, text_data):
        return json.loads(text_data)

    @classmethod
    def encode_json(cls, content):
        return json.dumps(content)



from django.conf import settings


def get(name, default=None):
    return getattr(settings, f'REACTOR_{name}', default)


LOGIN_URL = settings.LOGIN_URL
INCLUDE_TURBOLINKS = get('INCLUDE_TURBOLINKS', False)
USE_HTML_DIFF = get('USE_HTML_DIFF', True)
AUTO_BROADCAST = get('AUTO_BROADCAST', False)


if isinstance(AUTO_BROADCAST, bool):
    AUTO_BROADCAST = {
        # model_a
        # model_a.del
        # model_a.new
        'MODEL': AUTO_BROADCAST,

        # model_a.1234
        'MODEL_PK': AUTO_BROADCAST,

        # model_b.1234.model_a_set
        # model_b.1234.model_a_set.new
        # model_b.1234.model_a_set.del
        'RELATED': AUTO_BROADCAST,

        # model_b.1234.model_a_set
        # model_a.1234.model_b_set
        'M2M': AUTO_BROADCAST,
    }



from django.http.request import HttpRequest
from django.http.response import HttpResponse


def turbolinks_middleware(get_response):
    def middleware(request: HttpRequest):
        response: HttpResponse = get_response(request)
        response['Turbolinks-Location'] = request.get_full_path()
        return response
    return middleware
